@@ -8,6 +8,7 @@ from pydomo.groups import GroupClient
 from pydomo.pages import PageClient
 from pydomo.streams import StreamClient
 from pydomo.users import UserClient
+from pydomo.users import CreateUserRequest
 from pydomo.accounts import AccountClient
 from pydomo.utilities import UtilitiesClient
 from pandas import read_csv
@@ -15,6 +16,7 @@ from pandas import DataFrame
 from pandas import to_datetime
 from io import StringIO
 import logging
+import json
 
 DOMO = """####################################################################################################
 ####################################################################################################
@@ -79,11 +81,13 @@ class Domo:
         else:
             self.logger = parent_logger
 
+        timeout = kwargs.get('request_timeout', None)
+
         if kwargs.get('log_level'):
             self.logger.setLevel(kwargs['log_level'])
         self.logger.debug("\n" + DOMO + "\n")
 
-        self.transport = DomoAPITransport(client_id, client_secret, api_host, kwargs.get('use_https', True), self.logger)
+        self.transport = DomoAPITransport(client_id, client_secret, api_host, kwargs.get('use_https', True), self.logger, request_timeout = timeout)
         self.datasets = DataSetClient(self.transport, self.logger)
         self.groups = GroupClient(self.transport, self.logger)
         self.pages = PageClient(self.transport, self.logger)
@@ -105,14 +109,23 @@ class Domo:
         """
         return self.datasets.get(dataset_id)
 
-    def ds_delete(self, dataset_id):
+    def ds_delete(self, dataset_id, prompt_before_delete=True):
         """
-            Delete a DataSet naming convention equivilent with rDomo
+            Delete a DataSet naming convention equivalent with rdomo
             
             :Parameters:
             - `dataset_id`: id of a dataset (str)
         """
-        return self.datasets.delete(dataset_id)
+
+        del_data = 'Y'
+        if prompt_before_delete:
+            del_data = input("Permanently delete this data set? This is destructive and cannot be reversed. (Y/n)")
+
+        out = 'Data set not deleted'
+        if del_data == 'Y':
+            out = self.datasets.delete(dataset_id)
+
+        return out
 
     def ds_list(self, df_output = True, per_page=50, offset=0, limit=0):
         """
@@ -126,7 +139,7 @@ class Domo:
             - `per_page`:   results per page. Default 50 (int)
             - `offset`:     offset if you need to paginate results. Default 0 (int)
             - `limit`:      max ouput to return. If 0 then return all results on page. Default 0 (int)
-            
+
             :Returns:
             list or pandas dataframe depending on parameters
 
@@ -137,7 +150,7 @@ class Domo:
         else:
             out = DataFrame(list(l))
         return out
-    
+
     def ds_query(self, dataset_id, query, return_data=True):
         """
             Evaluate query and return dataset in a dataframe
@@ -150,7 +163,7 @@ class Domo:
             - `dataset_id`:     id of a dataset (str)
             - `query`:          query object (dict)
             - `return_data`:    should the result be a dataframe. Default True (Boolean)
-            
+
             :Returns:
             dict or pandas dataframe depending on parameters
         """
@@ -169,10 +182,10 @@ class Domo:
 
             :Parameters:
             - `dataset_id`:     id of a dataset (str)
-            
+
             :Returns:
             pandas dataframe
-        """    
+        """
         csv_download = self.datasets.data_export(dataset_id, include_csv_header=True)
 
         content = StringIO(csv_download)
@@ -185,21 +198,16 @@ class Domo:
                     df[col] = to_datetime(df[col])
                 except ValueError:
                     pass
+                except TypeError:
+                    pass
 
         return df
 
-    def ds_create(self, df_up, name, description=''):
-        dsr = DataSetRequest()
-        dsr.name = name
-        dsr.description = description
-        dsr.schema = Schema([Column(ColumnType.STRING, 'tt1'),
-                             Column(ColumnType.STRING, 'tt2')])
-
-        new_ds_info = self.datasets.create(dsr)
-
-        self.utilities.stream_upload(new_ds_info['id'],df_up,warn_schema_change=False)
-
-        return new_ds_info['id']
+    def ds_create(self, df_up, name, description='', update_method='REPLACE', key_column_names=''):
+        new_stream = self.utilities.stream_create(df_up, name, description, update_method, key_column_names)
+        ds_id = json.loads(new_stream.content.decode('utf-8'))['dataSet']['id']
+        self.utilities.stream_upload(ds_id,df_up,warn_schema_change=False)
+        return ds_id
 
 
     def ds_update(self, ds_id, df_up):
@@ -226,14 +234,14 @@ class Domo:
         ,   "not": false } ], "users": [ 27 ],"groups": [ ]}
 
         :Parameters:
-          - `dataset_id`:   id of the dataset PDP will be applied to (String) Required 
+          - `dataset_id`:   id of the dataset PDP will be applied to (String) Required
           Policy Object:
           - `name`: Name of the Policy (String) Required
-          - `filters[].column`:	Name of the column to filter on (String) Required 
+          - `filters[].column`:	Name of the column to filter on (String) Required
           - `filters[].not`: Determines if NOT is applied to the filter operation (Boolean) Required
-          - `filters[].operator`: Matching operator (EQUALS) (String) Required 
-          - `filters[].values[]`: Values to filter on (String) Required 
-          - `type`: Type of policy (user or system) (String) Required 
+          - `filters[].operator`: Matching operator (EQUALS) (String) Required
+          - `filters[].values[]`: Values to filter on (String) Required
+          - `type`: Type of policy (user or system) (String) Required
           - `users`: List of user IDs the policy applies to (array) Required
           - `groups`: List of group IDs the policy applies to (array) Required
         """
@@ -246,7 +254,7 @@ class Domo:
         >>> domo.pdp_delete('4405ff58-1957-45f0-82bd-914d989a3ea3', 35)
 
         :Parameters:
-        - `dataset_id`: id of the dataset PDP will be applied to (String) Required 
+        - `dataset_id`: id of the dataset PDP will be applied to (String) Required
         - `policy_id`:  id of the policy to delete (String) Required
         """
         return self.datasets.delete_pdp(dataset_id, policy_id)
@@ -261,7 +269,7 @@ class Domo:
             :Parameters:
             - `dataset_id`:   id of dataset with PDP policies (str) Required
             - `df_output`:  should the result be a dataframe. Default True (Boolean)
-            
+
             :Returns:
             list or pandas dataframe depending on parameters
 
@@ -291,15 +299,15 @@ class Domo:
         ,   "not": false } ], "users": [ 27 ],"groups": [ ]}
 
         :Parameters:
-          - `dataset_id`:   id of the dataset PDP will be applied to (String) Required 
-          - `policy_id`:    id of the PDP pollicy that will be updated (String) Required 
+          - `dataset_id`:   id of the dataset PDP will be applied to (String) Required
+          - `policy_id`:    id of the PDP pollicy that will be updated (String) Required
           Policy Object:
           - `name`: Name of the Policy (String) Required
-          - `filters[].column`:	Name of the column to filter on (String) Required 
+          - `filters[].column`:	Name of the column to filter on (String) Required
           - `filters[].not`: Determines if NOT is applied to the filter operation (Boolean) Required
-          - `filters[].operator`: Matching operator (EQUALS) (String) Required 
-          - `filters[].values[]`: Values to filter on (String) Required 
-          - `type`: Type of policy (user or system) (String) Required 
+          - `filters[].operator`: Matching operator (EQUALS) (String) Required
+          - `filters[].values[]`: Values to filter on (String) Required
+          - `type`: Type of policy (user or system) (String) Required
           - `users`: List of user IDs the policy applies to (array) Required
           - `groups`: List of group IDs the policy applies to (array) Required
         """
@@ -491,15 +499,27 @@ class Domo:
         """
             Add a User to a Group
         """
-        return self.groups.add_user(group_id, user_id)
+
+        if isinstance(user_id,list):
+            for x in user_id:
+                self.groups.add_user(group_id, x)
+        else:
+            self.groups.add_user(group_id, user_id)
+
+        return 'success'
 
 
 
-    def groups_create(self, group_request):
+    def groups_create(self, group_name, users=-1, active='true'):
         """
             Create a Group
         """
-        return self.groups.create(group_request)
+        req_body = {'name':group_name,'active':active}
+        grp_created = self.groups.create(req_body)
+        if (not isinstance(users,list) and users > 0) or isinstance(users,list):
+            self.groups_add_users(grp_created['id'],users)
+
+        return grp_created
 
 
 
@@ -507,6 +527,8 @@ class Domo:
         """
             Delete a Group
         """
+        existing_users = self.groups_list_users(group_id)
+        self.groups_remove_users(group_id,existing_users)
         return self.groups.delete(group_id)
 
 
@@ -519,19 +541,37 @@ class Domo:
 
 
 
-    def groups_list(self, limit=1000, offset=0):
+    def groups_list(self):
         """
             List all groups in Domo instance in a pandas dataframe.
         """
-        return DataFrame(self.groups.list(limit, offset))
+        grps = []
+        n_ret = 1
+        off = 0
+        batch_size = 500
+        while n_ret > 0:
+            gg = self.groups.list(batch_size,off*batch_size)
+            grps.extend(gg)
+            n_ret = gg.__len__()
+            off += 1
+        return DataFrame(grps)
 
 
-
-    def groups_list_users(self, group_id, limit=1000, offset=0):
+    def groups_list_users(self, group_id):
         """
             List Users in a Group
         """
-        return self.groups.list_users(group_id, limit, offset)
+        user_list = []
+        n_ret = 1
+        off = 0
+        batch_size=500
+        while n_ret > 0:
+            i_users = self.groups.list_users(group_id,limit=batch_size,offset=off*batch_size)
+            user_list.extend(i_users)
+            n_ret = i_users.__len__()
+            off += 1
+
+        return user_list
 
 
 
@@ -539,7 +579,14 @@ class Domo:
         """
             Remove a User to a Group
         """
-        return self.groups.remove_user(group_id, user_id)
+        if isinstance(user_id,list):
+            for x in user_id:
+                self.groups.remove_user(group_id, x)
+        else:
+            self.groups.remove_user(group_id, user_id)
+
+        return 'success'
+
 
 ######### Accounts #########
     def accounts_list(self):
@@ -555,7 +602,7 @@ class Domo:
         - `per_page`:   results per page. Default 50 (int)
         - `offset`:     offset if you need to paginate results. Default 0 (int)
         - `limit`:      max ouput to return. If 0 then return all results on page. Default 0 (int)
-            
+
 
         :returns:
           - A list of dicts (with nesting possible)
